@@ -95,7 +95,7 @@ void settings_timeout_cb(struct ev_loop *loop, ev_timer *w, int revents) {
     SSLOG(INFO, http2session) << "SETTINGS timeout";
   }
 
-  downstream_failure(http2session->get_addr());
+  downstream_failure(http2session->get_addr(), http2session->get_raddr());
 
   if (http2session->terminate_session(NGHTTP2_SETTINGS_TIMEOUT) != 0) {
     delete http2session;
@@ -111,6 +111,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto http2session = static_cast<Http2Session *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   if (LOG_ENABLED(INFO)) {
     SSLOG(INFO, http2session) << "Timeout";
   }
@@ -198,6 +202,7 @@ Http2Session::Http2Session(struct ev_loop *loop, SSL_CTX *ssl_ctx,
       group_(group),
       addr_(addr),
       session_(nullptr),
+      raddr_(nullptr),
       state_(DISCONNECTED),
       connection_check_state_(CONNECTION_CHECK_NONE),
       freelist_zone_(FREELIST_ZONE_NONE) {
@@ -240,6 +245,11 @@ int Http2Session::disconnect(bool hard) {
 
   wb_.reset();
 
+  if (dns_query_) {
+    auto dns_tracker = worker_->get_dns_tracker();
+    dns_tracker->cancel(dns_query_.get());
+  }
+
   conn_.rlimit.stopw();
   conn_.wlimit.stopw();
 
@@ -298,12 +308,47 @@ int Http2Session::disconnect(bool hard) {
   return 0;
 }
 
+int Http2Session::resolve_name() {
+  int rv;
+
+  auto dns_query = make_unique<DNSQuery>(
+      addr_->host, [this](int status, const Address *result) {
+        int rv;
+
+        if (status == DNS_STATUS_OK) {
+          *resolved_addr_ = *result;
+          util::set_port(*this->resolved_addr_, this->addr_->port);
+        }
+
+        rv = this->initiate_connection();
+        if (rv != 0) {
+          delete this;
+        }
+      });
+  resolved_addr_ = make_unique<Address>();
+  auto dns_tracker = worker_->get_dns_tracker();
+  rv = dns_tracker->resolve(resolved_addr_.get(), dns_query.get());
+  switch (rv) {
+  case DNS_STATUS_ERROR:
+    return -1;
+  case DNS_STATUS_RUNNING:
+    dns_query_ = std::move(dns_query);
+    state_ = RESOLVING_NAME;
+    return 0;
+  case DNS_STATUS_OK:
+    util::set_port(*resolved_addr_, addr_->port);
+    return 0;
+  default:
+    assert(0);
+  }
+}
+
 int Http2Session::initiate_connection() {
   int rv = 0;
 
   auto worker_blocker = worker_->get_connect_blocker();
 
-  if (state_ == DISCONNECTED) {
+  if (state_ == DISCONNECTED || state_ == RESOLVING_NAME) {
     if (worker_blocker->blocked()) {
       if (LOG_ENABLED(INFO)) {
         SSLOG(INFO, this)
@@ -313,6 +358,8 @@ int Http2Session::initiate_connection() {
     }
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
   const auto &proxy = get_config()->downstream_http_proxy;
   if (!proxy.host.empty() && state_ == DISCONNECTED) {
     if (LOG_ENABLED(INFO)) {
@@ -344,6 +391,8 @@ int Http2Session::initiate_connection() {
       return -1;
     }
 
+    raddr_ = &proxy.addr;
+
     worker_blocker->on_success();
 
     ev_io_set(&conn_.rev, conn_.fd, EV_READ);
@@ -351,7 +400,7 @@ int Http2Session::initiate_connection() {
 
     conn_.wlimit.startw();
 
-    // TODO we should have timeout for connection establishment
+    conn_.wt.repeat = downstreamconf.timeout.connect;
     ev_timer_again(conn_.loop, &conn_.wt);
 
     write_ = &Http2Session::connected;
@@ -368,36 +417,68 @@ int Http2Session::initiate_connection() {
     return 0;
   }
 
-  if (state_ == DISCONNECTED || state_ == PROXY_CONNECTED) {
+  if (state_ == DISCONNECTED || state_ == PROXY_CONNECTED ||
+      state_ == RESOLVING_NAME) {
     if (LOG_ENABLED(INFO)) {
-      SSLOG(INFO, this) << "Connecting to downstream server";
+      if (state_ != RESOLVING_NAME) {
+        SSLOG(INFO, this) << "Connecting to downstream server";
+      }
     }
     if (addr_->tls) {
       assert(ssl_ctx_);
 
-      auto ssl = ssl::create_ssl(ssl_ctx_);
-      if (!ssl) {
-        return -1;
+      if (state_ != RESOLVING_NAME) {
+        auto ssl = ssl::create_ssl(ssl_ctx_);
+        if (!ssl) {
+          return -1;
+        }
+
+        ssl::setup_downstream_http2_alpn(ssl);
+
+        conn_.set_ssl(ssl);
+
+        auto sni_name =
+            addr_->sni.empty() ? StringRef{addr_->host} : StringRef{addr_->sni};
+
+        if (!util::numeric_host(sni_name.c_str())) {
+          // TLS extensions: SNI. There is no documentation about the return
+          // code for this function (actually this is macro wrapping SSL_ctrl
+          // at the time of this writing).
+          SSL_set_tlsext_host_name(conn_.tls.ssl, sni_name.c_str());
+        }
+
+        auto tls_session = ssl::reuse_tls_session(addr_->tls_session_cache);
+        if (tls_session) {
+          SSL_set_session(conn_.tls.ssl, tls_session);
+          SSL_SESSION_free(tls_session);
+        }
       }
 
-      ssl::setup_downstream_http2_alpn(ssl);
-
-      conn_.set_ssl(ssl);
-
-      auto sni_name =
-          addr_->sni.empty() ? StringRef{addr_->host} : StringRef{addr_->sni};
-
-      if (!util::numeric_host(sni_name.c_str())) {
-        // TLS extensions: SNI. There is no documentation about the return
-        // code for this function (actually this is macro wrapping SSL_ctrl
-        // at the time of this writing).
-        SSL_set_tlsext_host_name(conn_.tls.ssl, sni_name.c_str());
+      if (state_ == DISCONNECTED) {
+        if (addr_->dns) {
+          rv = resolve_name();
+          if (rv != 0) {
+            downstream_failure(addr_, nullptr);
+            return -1;
+          }
+          if (state_ == RESOLVING_NAME) {
+            return 0;
+          }
+          raddr_ = resolved_addr_.get();
+        } else {
+          raddr_ = &addr_->addr;
+        }
       }
 
-      auto tls_session = ssl::reuse_tls_session(addr_->tls_session_cache);
-      if (tls_session) {
-        SSL_set_session(conn_.tls.ssl, tls_session);
-        SSL_SESSION_free(tls_session);
+      if (state_ == RESOLVING_NAME) {
+        if (dns_query_->status == DNS_STATUS_ERROR) {
+          downstream_failure(addr_, nullptr);
+          return -1;
+        }
+        assert(dns_query_->status == DNS_STATUS_OK);
+        state_ = DISCONNECTED;
+        dns_query_.reset();
+        raddr_ = resolved_addr_.get();
       }
 
       // If state_ == PROXY_CONNECTED, we has connected to the proxy
@@ -405,12 +486,11 @@ int Http2Session::initiate_connection() {
       if (state_ == DISCONNECTED) {
         assert(conn_.fd == -1);
 
-        conn_.fd =
-            util::create_nonblock_socket(addr_->addr.su.storage.ss_family);
+        conn_.fd = util::create_nonblock_socket(raddr_->su.storage.ss_family);
         if (conn_.fd == -1) {
           auto error = errno;
           SSLOG(WARN, this)
-              << "socket() failed; addr=" << util::to_numeric_addr(&addr_->addr)
+              << "socket() failed; addr=" << util::to_numeric_addr(raddr_)
               << ", errno=" << error;
 
           worker_blocker->on_failure();
@@ -421,15 +501,14 @@ int Http2Session::initiate_connection() {
 
         rv = connect(conn_.fd,
                      // TODO maybe not thread-safe?
-                     const_cast<sockaddr *>(&addr_->addr.su.sa),
-                     addr_->addr.len);
+                     const_cast<sockaddr *>(&raddr_->su.sa), raddr_->len);
         if (rv != 0 && errno != EINPROGRESS) {
           auto error = errno;
-          SSLOG(WARN, this) << "connect() failed; addr="
-                            << util::to_numeric_addr(&addr_->addr)
-                            << ", errno=" << error;
+          SSLOG(WARN, this)
+              << "connect() failed; addr=" << util::to_numeric_addr(raddr_)
+              << ", errno=" << error;
 
-          downstream_failure(addr_);
+          downstream_failure(addr_, raddr_);
           return -1;
         }
 
@@ -439,17 +518,44 @@ int Http2Session::initiate_connection() {
 
       conn_.prepare_client_handshake();
     } else {
+      if (state_ == DISCONNECTED) {
+        // Without TLS and proxy.
+        if (addr_->dns) {
+          rv = resolve_name();
+          if (rv != 0) {
+            downstream_failure(addr_, nullptr);
+            return -1;
+          }
+          if (state_ == RESOLVING_NAME) {
+            return 0;
+          }
+          raddr_ = resolved_addr_.get();
+        } else {
+          raddr_ = &addr_->addr;
+        }
+      }
+
+      if (state_ == RESOLVING_NAME) {
+        if (dns_query_->status == DNS_STATUS_ERROR) {
+          downstream_failure(addr_, nullptr);
+          return -1;
+        }
+        assert(dns_query_->status == DNS_STATUS_OK);
+        state_ = DISCONNECTED;
+        dns_query_.reset();
+        raddr_ = resolved_addr_.get();
+      }
+
       if (state_ == DISCONNECTED) {
         // Without TLS and proxy.
         assert(conn_.fd == -1);
 
-        conn_.fd =
-            util::create_nonblock_socket(addr_->addr.su.storage.ss_family);
+        conn_.fd = util::create_nonblock_socket(raddr_->su.storage.ss_family);
 
         if (conn_.fd == -1) {
           auto error = errno;
           SSLOG(WARN, this)
-              << "socket() failed; addr=" << util::to_numeric_addr(&addr_->addr)
+              << "socket() failed; addr=" << util::to_numeric_addr(raddr_)
               << ", errno=" << error;
 
           worker_blocker->on_failure();
@@ -458,15 +564,15 @@ int Http2Session::initiate_connection() {
 
         worker_blocker->on_success();
 
-        rv = connect(conn_.fd, const_cast<sockaddr *>(&addr_->addr.su.sa),
-                     addr_->addr.len);
+        rv = connect(conn_.fd, const_cast<sockaddr *>(&raddr_->su.sa),
+                     raddr_->len);
         if (rv != 0 && errno != EINPROGRESS) {
           auto error = errno;
-          SSLOG(WARN, this) << "connect() failed; addr="
-                            << util::to_numeric_addr(&addr_->addr)
-                            << ", errno=" << error;
+          SSLOG(WARN, this)
+              << "connect() failed; addr=" << util::to_numeric_addr(raddr_)
+              << ", errno=" << error;
 
-          downstream_failure(addr_);
+          downstream_failure(addr_, raddr_);
           return -1;
         }
 
@@ -475,25 +581,22 @@ int Http2Session::initiate_connection() {
       }
     }
 
-    write_ = &Http2Session::connected;
-
     on_write_ = &Http2Session::downstream_write;
     on_read_ = &Http2Session::downstream_read;
 
     // We have been already connected when no TLS and proxy is used.
-    if (state_ != CONNECTED) {
-      state_ = CONNECTING;
-      conn_.wlimit.startw();
-      // TODO we should have timeout for connection establishment
-      ev_timer_again(conn_.loop, &conn_.wt);
-    } else {
-      conn_.rlimit.startw();
-
-      if (addr_->num_dconn == 0) {
-        ev_timer_again(conn_.loop, &conn_.rt);
-      }
+    if (state_ == PROXY_CONNECTED) {
+      return connected();
     }
 
+    write_ = &Http2Session::connected;
+
+    state_ = CONNECTING;
+    conn_.wlimit.startw();
+
+    conn_.wt.repeat = downstreamconf.timeout.connect;
+    ev_timer_again(conn_.loop, &conn_.wt);
+
     return 0;
   }
 
@@ -612,8 +715,6 @@ int Http2Session::downstream_connect_proxy() {
 void Http2Session::add_downstream_connection(Http2DownstreamConnection *dconn) {
   dconns_.append(dconn);
   ++addr_->num_dconn;
-
-  stop_read_timer();
 }
 
 void Http2Session::remove_downstream_connection(
@@ -622,10 +723,6 @@ void Http2Session::remove_downstream_connection(
   dconns_.remove(dconn);
   dconn->detach_stream_data();
 
-  if (addr_->num_dconn == 0) {
-    repeat_read_timer();
-  }
-
   if (LOG_ENABLED(INFO)) {
     SSLOG(INFO, this) << "Remove downstream";
   }
@@ -860,7 +957,7 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
         nghttp2_session_get_stream_user_data(session, promised_stream_id));
     if (!promised_sd || !promised_sd->dconn) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
-      return 0;
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     auto promised_downstream = promised_sd->dconn->get_downstream();
@@ -965,7 +1062,7 @@ int on_begin_headers_callback(nghttp2_session *session,
         nghttp2_session_get_stream_user_data(session, frame->hd.stream_id));
     if (!sd || !sd->dconn) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
-      return 0;
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     auto downstream = sd->dconn->get_downstream();
@@ -976,6 +1073,7 @@ int on_begin_headers_callback(nghttp2_session *session,
     if (http2session->handle_downstream_push_promise(downstream,
                                                      promised_stream_id) != 0) {
       http2session->submit_rst_stream(promised_stream_id, NGHTTP2_CANCEL);
+      return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
     }
 
     return 0;
@@ -1054,8 +1152,8 @@ int on_response_headers(Http2Session *http2session, Downstream *downstream,
     }
     downstream->set_request_state(Downstream::HEADER_COMPLETE);
     if (LOG_ENABLED(INFO)) {
-      SSLOG(INFO, http2session)
-          << "HTTP upgrade success. stream_id=" << frame->hd.stream_id;
+      SSLOG(INFO, http2session) << "HTTP upgrade success. stream_id="
+                                << frame->hd.stream_id;
     }
   } else {
     auto content_length = resp.fs.header(http2::HD_CONTENT_LENGTH);
@@ -1546,7 +1644,7 @@ int Http2Session::connection_made() {
 #endif // OPENSSL_VERSION_NUMBER >= 0x10002000L
 
     if (!next_proto) {
-      downstream_failure(addr_);
+      downstream_failure(addr_, raddr_);
       return -1;
     }
 
@@ -1555,12 +1653,13 @@ int Http2Session::connection_made() {
       SSLOG(INFO, this) << "Negotiated next protocol: " << proto;
     }
     if (!util::check_h2_is_selected(proto)) {
-      downstream_failure(addr_);
+      downstream_failure(addr_, raddr_);
       return -1;
     }
   }
 
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &http2conf = config->http2;
 
   rv = nghttp2_session_client_new2(&session_, http2conf.downstream.callbacks,
                                    this, http2conf.downstream.option);
@@ -1577,7 +1676,7 @@ int Http2Session::connection_made() {
   entry[1].settings_id = NGHTTP2_SETTINGS_INITIAL_WINDOW_SIZE;
   entry[1].value = http2conf.downstream.window_size;
 
-  if (http2conf.no_server_push || get_config()->http2_proxy) {
+  if (http2conf.no_server_push || config->http2_proxy) {
     entry[nentry].settings_id = NGHTTP2_SETTINGS_ENABLE_PUSH;
     entry[nentry].value = 0;
     ++nentry;
@@ -1843,10 +1942,10 @@ int Http2Session::connected() {
   auto sock_error = util::get_socket_error(conn_.fd);
   if (sock_error != 0) {
     SSLOG(WARN, this) << "Backend connect failed; addr="
-                      << util::to_numeric_addr(&addr_->addr)
+                      << util::to_numeric_addr(raddr_)
                       << ": errno=" << sock_error;
 
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1855,7 +1954,14 @@ int Http2Session::connected() {
     SSLOG(INFO, this) << "Connection established";
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  // Reset timeout for write.  Previously, we set timeout for connect.
+  conn_.wt.repeat = downstreamconf.timeout.write;
+  ev_timer_again(conn_.loop, &conn_.wt);
+
   conn_.rlimit.startw();
+  conn_.again_rt();
 
   read_ = &Http2Session::read_clear;
   write_ = &Http2Session::write_clear;
@@ -1880,6 +1986,8 @@ int Http2Session::connected() {
 }
 
 int Http2Session::read_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<uint8_t, 16_k> buf;
 
   for (;;) {
@@ -1900,6 +2008,8 @@ int Http2Session::read_clear() {
 }
 
 int Http2Session::write_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<struct iovec, MAX_WR_IOVCNT> iov;
 
   for (;;) {
@@ -1912,7 +2022,12 @@ int Http2Session::write_clear() {
       }
 
       if (nwrite < 0) {
-        return nwrite;
+        // We may have pending data in receive buffer which may
+        // contain part of response body.  So keep reading.  Invoke
+        // read event to get read(2) error just in case.
+        ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+        write_ = &Http2Session::write_void;
+        break;
       }
 
       wb_.drain(nwrite);
@@ -1934,7 +2049,7 @@ int Http2Session::write_clear() {
 }
 
 int Http2Session::tls_handshake() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   ERR_clear_error();
 
@@ -1945,7 +2060,7 @@ int Http2Session::tls_handshake() {
   }
 
   if (rv < 0) {
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return rv;
   }
@@ -1955,8 +2070,8 @@ int Http2Session::tls_handshake() {
   }
 
   if (!get_config()->tls.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_) != 0) {
-    downstream_failure(addr_);
+      ssl::check_cert(conn_.tls.ssl, addr_, raddr_) != 0) {
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1964,8 +2079,8 @@ int Http2Session::tls_handshake() {
   if (!SSL_session_reused(conn_.tls.ssl)) {
     auto tls_session = SSL_get0_session(conn_.tls.ssl);
     if (tls_session) {
-      ssl::try_cache_tls_session(addr_->tls_session_cache, addr_->addr,
-                                 tls_session, ev_now(conn_.loop));
+      ssl::try_cache_tls_session(addr_->tls_session_cache, *raddr_, tls_session,
+                                 ev_now(conn_.loop));
     }
   }
 
@@ -1981,6 +2096,8 @@ int Http2Session::tls_handshake() {
 }
 
 int Http2Session::read_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<uint8_t, 16_k> buf;
 
   ERR_clear_error();
@@ -2003,6 +2120,8 @@ int Http2Session::read_tls() {
 }
 
 int Http2Session::write_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
   struct iovec iov;
@@ -2018,7 +2137,12 @@ int Http2Session::write_tls() {
       }
 
       if (nwrite < 0) {
-        return nwrite;
+        // We may have pending data in receive buffer which may
+        // contain part of response body.  So keep reading.  Invoke
+        // read event to get read(2) error just in case.
+        ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+        write_ = &Http2Session::write_void;
+        break;
       }
 
       wb_.drain(nwrite);
@@ -2041,6 +2165,11 @@ int Http2Session::write_tls() {
   return 0;
 }
 
+int Http2Session::write_void() {
+  conn_.wlimit.stopw();
+  return 0;
+}
+
 bool Http2Session::should_hard_fail() const {
   switch (state_) {
   case PROXY_CONNECTING:
@@ -2248,9 +2377,9 @@ void Http2Session::on_timeout() {
   }
   case CONNECTING: {
     SSLOG(WARN, this) << "Connect time out; addr="
-                      << util::to_numeric_addr(&addr_->addr);
+                      << util::to_numeric_addr(raddr_);
 
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
     break;
   }
   }
@@ -2270,10 +2399,6 @@ void Http2Session::check_retire() {
   signal_write();
 }
 
-void Http2Session::repeat_read_timer() {
-  ev_timer_again(conn_.loop, &conn_.rt);
-}
-
-void Http2Session::stop_read_timer() { ev_timer_stop(conn_.loop, &conn_.rt); }
+const Address *Http2Session::get_raddr() const { return raddr_; }
 
 } // namespace shrpx
