@@ -42,6 +42,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto mconn = static_cast<MemcachedConnection *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   if (LOG_ENABLED(INFO)) {
     MCLOG(INFO, mconn) << "Time out";
   }
@@ -100,7 +104,7 @@ MemcachedConnection::MemcachedConnection(const Address *addr,
             connectcb, readcb, timeoutcb, this, 0, 0., PROTO_MEMCACHED),
       do_read_(&MemcachedConnection::noop),
       do_write_(&MemcachedConnection::noop),
-      sni_name_(sni_name.str()),
+      sni_name_(sni_name),
       connect_blocker_(gen, loop, [] {}, [] {}),
       parse_state_{},
       addr_(addr),
@@ -225,6 +229,8 @@ int MemcachedConnection::connected() {
   ev_set_cb(&conn_.wev, writecb);
 
   if (conn_.tls.ssl) {
+    conn_.again_rt();
+
     do_read_ = &MemcachedConnection::tls_handshake;
     do_write_ = &MemcachedConnection::tls_handshake;
 
@@ -249,7 +255,7 @@ int MemcachedConnection::on_read() { return do_read_(*this); }
 int MemcachedConnection::tls_handshake() {
   ERR_clear_error();
 
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   auto rv = conn_.tls_handshake();
   if (rv == SHRPX_ERR_INPROGRESS) {
@@ -268,7 +274,7 @@ int MemcachedConnection::tls_handshake() {
   auto &tlsconf = get_config()->tls;
 
   if (!tlsconf.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_, StringRef(sni_name_)) != 0) {
+      ssl::check_cert(conn_.tls.ssl, addr_, sni_name_) != 0) {
     connect_blocker_.on_failure();
     return -1;
   }
@@ -299,12 +305,7 @@ int MemcachedConnection::write_tls() {
     return 0;
   }
 
-  if (sendq_.empty()) {
-    conn_.wlimit.stopw();
-    ev_timer_stop(conn_.loop, &conn_.wt);
-
-    return 0;
-  }
+  conn_.last_read = ev_now(conn_.loop);
 
   std::array<struct iovec, MAX_WR_IOVCNT> iov;
   std::array<uint8_t, 16_k> buf;
@@ -343,9 +344,7 @@ int MemcachedConnection::read_tls() {
     return 0;
   }
 
-  if (ev_is_active(&conn_.rt)) {
-    ev_timer_again(conn_.loop, &conn_.rt);
-  }
+  conn_.last_read = ev_now(conn_.loop);
 
   for (;;) {
     auto nread = conn_.read_tls(recvbuf_.last, recvbuf_.wleft());
@@ -373,12 +372,7 @@ int MemcachedConnection::write_clear() {
     return 0;
   }
 
-  if (sendq_.empty()) {
-    conn_.wlimit.stopw();
-    ev_timer_stop(conn_.loop, &conn_.wt);
-
-    return 0;
-  }
+  conn_.last_read = ev_now(conn_.loop);
 
   std::array<struct iovec, MAX_WR_IOVCNT> iov;
 
@@ -406,9 +400,7 @@ int MemcachedConnection::read_clear() {
     return 0;
   }
 
-  if (ev_is_active(&conn_.rt)) {
-    ev_timer_again(conn_.loop, &conn_.rt);
-  }
+  conn_.last_read = ev_now(conn_.loop);
 
   for (;;) {
     auto nread = conn_.read_clear(recvbuf_.last, recvbuf_.wleft());
@@ -552,8 +544,8 @@ int MemcachedConnection::parse_packet() {
 
       if (LOG_ENABLED(INFO)) {
         if (parse_state_.status_code) {
-          MCLOG(INFO, this)
-              << "response returned error status: " << parse_state_.status_code;
+          MCLOG(INFO, this) << "response returned error status: "
+                            << parse_state_.status_code;
         }
       }
 
@@ -669,7 +661,7 @@ void MemcachedConnection::drain_send_queue(size_t nwrite) {
   if (recvq_.empty()) {
     ev_timer_stop(conn_.loop, &conn_.rt);
   } else if (!ev_is_active(&conn_.rt)) {
-    ev_timer_again(conn_.loop, &conn_.rt);
+    conn_.again_rt();
   }
 }
 
